import json
from nimoy.utils import import_class, fingerprint as _fp, ensure_key_order
from collections import OrderedDict


class DocumentSchema(object):

    def __init__(self, fields, indexes, **options):
        self._fields = OrderedDict([(fname, self._parse_field(fops)) for fname, fops in fields.items()])
        self._indexes = [self._parse_index(index) for index in indexes]
        self._options = options

    @classmethod
    def from_json(cls, json_string=None):
        json_document = json.loads(json_string)
        return cls.from_dict(json_document)

    @classmethod
    def from_dict(cls, schema_dict):
        return cls(fields=schema_dict.get('fields', {}),
                   indexes=schema_dict.get('indexes', []),
                   **schema_dict.get('options', {}))

    def to_dict(self):
        return ensure_key_order({
            'fields': {fname: self._field_ops_to_dict(fops) for fname, fops in self._fields.items()},
            'indexes': self._indexes,
            'options': self._options
        })

    def to_json(self):
        return json.dumps(self.to_dict(), sort_keys=True)

    def _parse_field(self, options):
        return {
            'type': import_class(options.get('type', 'TextField'), default_prefix='nimoy.fields'),
            'required': options.get('required', False)
        }

    def _field_ops_to_dict(self, options):
        return {
            'type': options['type'].to_text(),
            'required': options['required']
        }

    def _parse_index(self, index):
        def _is_valid_field(field):
            return field in self._fields
        if all([_is_valid_field(field) for field in index]):
            return index
        else:  # pragma: nocover
            raise Exception("Invalid index {}".format(str(index)))

    @property
    def fingerprint(self):
        return _fp(self.to_json())

    @property
    def primary_key(self):
        return self._indexes[0]

    def validate_data(self, _data):
        return _data


class DBSchema(object):

    def __init__(self, schemas=None, **options):
        schemas = schemas or {}
        self._schemas = {name: self._parse_schema_dict(schema_dict) for name, schema_dict in schemas.items()}
        self._options = options

    @classmethod
    def from_json(cls, json_string=None):
        json_document = json.loads(json_string)
        return cls.from_dict(json_document)

    @classmethod
    def from_dict(cls, schema_dict, **options):
        return cls(schemas=schema_dict, **options)

    def _parse_schema_dict(self, schema_dict):
        return DocumentSchema(fields=schema_dict['fields'], indexes=schema_dict['indexes'], **schema_dict.get('options', {}))

    def to_json(self):
        return json.dumps(self.to_dict(), sort_keys=True)

    def to_dict(self):
        return ensure_key_order({
            'schemas': {sname: schema.to_dict() for sname, schema in self._schemas.items()},
            'options': self._options
        })

    def validate_data(self, schema_name, _data):
        if schema_name in self._schemas:
            return self._schemas[schema_name].validate_data(_data)
        raise ValueError("Invalid data for {}: {}".format(schema_name, _data))

    def get_fingerprint(self, schema_name):
        return self._schemas[schema_name].fingerprint

    def get_primary_key(self, schema_name):
        return self._schemas[schema_name].primary_key



from .base import BaseBackend
from nimoy.exceptions import ItemNotFound
from unicodedata import normalize
import operator


_global_state = {}


class InMemoryBackend(BaseBackend):
    def _primary_key(self, schema_name, _data):
        key_names = self._conn.schema.get_primary_key(schema_name)
        ret = []
        for key in key_names:
            ret.append(_data.get(key))
        if len(ret) == 1:
            return ret[0]
        else:
            return tuple(ret)

    def put_item(self, schema_name, _data, overwrite=False):
        if schema_name not in _global_state:
            _global_state[schema_name] = {}
        _global_state[schema_name][self._primary_key(schema_name, _data)] = _data
        return True

    def get_item(self, schema_name, _id):
        try:
            return _global_state[schema_name].get(_id)
        except KeyError:
            raise ItemNotFound("Item not found for id {} in {}.".format(_id, schema_name))

    def delete_item(self, schema_name, _id):
        if _id in _global_state[schema_name]:
            del _global_state[schema_name][_id]
        return True

    def _query_or_scan(self, schema_name, _w, limit=10):
        predicate = parse_wt(_w)
        return (v for k, v in _global_state.get(schema_name, {}).items() if predicate(v))

    def query(self, schema_name, _w, limit=10):
        return self._query_or_scan(schema_name, _w, limit)

    def scan(self, schema_name, _w, limit=10):
        return self._query_or_scan(schema_name, _w, limit)

    def query_count(self, schema_name, _w):
        return len(list(self._query_or_scan(schema_name, _w)))


def parse_wt(wt, fmap=None):
    _fmap = fmap or {}

    def _x_f(obj, field):
        if field in _fmap:
            return _fmap[field](obj[field])
        else:
            return obj.get(field)

    def _get_normalized_values(obj, field, value):
        v_normalized = value
        if hasattr(obj[field], 'lower'):
            cobj = _copy(obj)
            cobj[field] = _normalize_and_lower(cobj[field])
            v_normalized = _normalize_and_lower(v_normalized)
            return cobj, v_normalized
        return obj, v_normalized

    def _eq(field, value):
        def _filter(obj):
            return _x_f(obj, field) == value

        def _filter_embed(obj):
            emb_key, emb_field = field.split('.')
            return value in [_x_f(d, emb_field) for d in obj.get(emb_key, [])]
        return _filter_embed if '.' in field else _filter

    def _field_value_predicate(field, value, _operator):
        return lambda obj: _operator(_x_f(obj, field), value)

    def _value_field_predicate(field, value, _operator):
        return lambda obj: _operator(value, _x_f(obj, field))

    def _missing(field, value):
        return lambda obj: operator.is_(_x_f(obj, field), None)

    def _contains(field, value):
        def _filter(obj):
            if obj[field] is not None:
                obj_normalized, v_normalized = _get_normalized_values(obj, field, value)
                if hasattr(v_normalized, 'split'):
                    return all([v_n in _x_f(obj_normalized, field) for v_n in v_normalized.split(" ")])
                return v_normalized in _x_f(obj_normalized, field)
            return False
        return _filter

    def _ncontains(field, value):
        def _filter(obj):
            if obj[field] is not None:
                obj_normalized, v_normalized = _get_normalized_values(obj, field, value)
                if hasattr(v_normalized, 'split'):
                    return not all([v_n in _x_f(obj_normalized, field) for v_n in v_normalized.split(" ")])
                return v_normalized not in _x_f(obj_normalized, field)
            return False
        return _filter

    def _in(field, value):
        return lambda obj: operator.contains(value, _x_f(obj, field))

    def _nin(field, value):
        return lambda obj: operator.not_(operator.contains(value, _x_f(obj, field)))

    def _and(wts):
        _filters = [parse_wt(_wt, fmap=fmap) for _wt in wts]

        def _filter(obj):
            return all([_f(obj) for _f in _filters])
        return _filter

    def _or(wts):
        _filters = [parse_wt(_wt, fmap=fmap) for _wt in wts]

        def _filter(obj):
            return any([_f(obj) for _f in _filters])
        return _filter

    def _startswith(field, value):
        return lambda obj: _x_f(obj, field).startswith(value) if hasattr(_x_f(obj, field), 'startswith') else False

    if wt[0] == 'eq':
        return _eq(wt[1], wt[2])
    elif wt[0] == 'neq':
        return _field_value_predicate(wt[1], wt[2], _operator=operator.ne)
    elif wt[0] == 'gt':
        return _field_value_predicate(wt[1], wt[2], _operator=operator.gt)
    elif wt[0] == 'gte':
        return _field_value_predicate(wt[1], wt[2], _operator=operator.ge)
    elif wt[0] == 'lt':
        return _field_value_predicate(wt[1], wt[2], _operator=operator.lt)
    elif wt[0] == 'lte':
        return _field_value_predicate(wt[1], wt[2], _operator=operator.le)
    elif wt[0] == 'contains':
        return _contains(wt[1], wt[2])
    elif wt[0] == 'ncontains':
        return _ncontains(wt[1], wt[2])
    elif wt[0] == 'in':
        return _in(wt[1], wt[2])
    elif wt[0] == 'nin':
        return _nin(wt[1], wt[2])
    elif wt[0] == 'and':
        return _and(wt[1:])
    elif wt[0] == 'or':
        return _or(wt[1:])
    elif wt[0] == 'startswith':
        return _startswith(wt[1], wt[2])
    elif wt[0] == 'missing':
        return _missing(wt[1], wt[2])


def _copy(d):
    def _copy_list(l):
        r = []
        for i in l:
            if isinstance(i, dict):
                r.append(_copy_dict(i))
            elif isinstance(i, list):
                r.append(_copy_list(i))
            else:
                r.append(i)
        return r

    def _copy_dict(d):
        r = {}
        for i, j in d.items():
            if isinstance(j, dict):
                r[i] = _copy_dict(j)
            elif isinstance(j, list):
                r[i] = _copy_list(j)
            else:
                r[i] = j
        return r

    """ Create a new dict from d """
    return _copy_dict(d)


def _normalize_and_lower(value):
    return normalize('NFKD', value.lower())\
        .encode('ascii', 'ignore')\
        .decode('utf-8')



import decimal
from nimoy.exceptions import ItemNotFound
from .base import BaseBackend

try:
    from boto.dynamodb2.table import Table
    from boto.dynamodb2.exceptions import ItemNotFound as BotoItemNotFound
except ImportError:
    Table = None
    BotoItemNotFound = None


class DynamoDBBackend(BaseBackend):
    def __init__(self, conn):
        if Table is None:
            raise ImportError("Plz. install boto library for DynamoDBBackend.")
        super().__init__(conn)

    def _kwargs_for_id(self, schema_name, _id):
        key_names = self._conn.schema.get_primary_key(schema_name)
        if len(key_names) > 1:
            kwargs = {name: _id[idx] for idx, name in enumerate(key_names)}
        else:
            kwargs = {key_names[0]: _id}
        return kwargs

    def put_item(self, schema_name, _data, overwrite=False):
        table = Table(schema_name)
        return table.put_item(data=_data, overwrite=overwrite)

    def get_item(self, schema_name, _id):
        try:
            table = Table(schema_name)
            return table.get_item(**self._kwargs_for_id(schema_name, _id))
        except BotoItemNotFound:
            raise ItemNotFound("Item not found for id {} in {}.".format(_id, schema_name))

    def delete_item(self, schema_name, _id):
        table = Table(schema_name)
        return table.delete_item(**self._kwargs_for_id(schema_name, _id))

    def query(self, schema_name, _w, limit=10):
        table = Table(schema_name)
        kwargs = parse_w_for_kwargs(_w)
        kwargs['limit'] = limit
        return table.query_2(**kwargs)

    def scan(self, schema_name, _w, limit=10):
        table = Table(schema_name)
        kwargs = parse_w_for_kwargs(_w)
        kwargs['limit'] = limit
        return table.scan(**kwargs)

    def query_count(self, schema_name, _w):
        table = Table(schema_name)
        kwargs = parse_w_for_kwargs(_w)
        return table.query_count(**kwargs)

    def prepare_data(self, schema_name, _data, level=0):
        if isinstance(_data, dict):
            return {k: self.prepare_data(schema_name, v, level=level+1) for k, v in _data.items()}
        elif isinstance(_data, (str, int)):
            return _data
        elif isinstance(_data, float):
            return decimal.Decimal(str(_data))
        elif isinstance(_data, (list, tuple)):
            return [self.prepare_data(schema_name, v, level=level+1) for v in _data]
        else:
            return _data


def parse_w_for_kwargs(_w):
    ret = {}
    if _w[0] == 'and':
        ret.update(parse_w_for_kwargs(_w[1]))
        ret.update(parse_w_for_kwargs(_w[2]))
    elif _w[0] in ['eq', 'beginswith', 'lte', 'gte', 'lt', 'gt']:  # FIXME
        ret['__'.join([_w[1], _w[0]])] = _w[2]
    return ret



import importlib
import hashlib
from collections import OrderedDict


def import_class(cls_path, default_prefix='nimoy'):
    if isinstance(cls_path, str):
        if '.' not in cls_path:
            cls_path = '.'.join([default_prefix, cls_path])
        dotposition = cls_path.rfind('.')
        module_name = cls_path[:dotposition]
        class_name = cls_path[dotposition + 1:]
        mod = importlib.import_module(module_name)
        return getattr(mod, class_name)
    else:
        return cls_path


def fingerprint(_str):
    sha = hashlib.sha1()
    sha.update(_str.encode('utf-8'))
    return sha.hexdigest()


def ensure_key_order(dictionary):
    ret = OrderedDict()
    for key in sorted(dictionary):
        if isinstance(dictionary[key], dict):
            ret[key] = ensure_key_order(dictionary[key])
        else:
            ret[key] = dictionary[key]
    return ret



class BaseBackend(object):
    def __init__(self, conn):
        self._conn = conn

    def prepare_data(self, schema_name, _data):
        return _data

    def put_item(self, schema_name, _data):
        raise NotImplemented()

    def get_item(self, schema_name, _id):
        raise NotImplemented()

    def delete_item(self, schema_name, _id):
        raise NotImplemented()



class BaseScalarField(object):

    @classmethod
    def to_text(cls):
        return '.'.join(['nimoy.fields', cls.__name__])


class IntegerField(BaseScalarField):
    pass


class TextField(BaseScalarField):
    pass


class DatetimeField(BaseScalarField):
    pass


class IDField(BaseScalarField):
    pass



# coding: utf-8
from setuptools import setup


__version__ = None
exec(open('nimoy/version.py').read())


def next_version():
    _v = __version__.split('.')
    _v[-1] = str(int(_v[-1]) + 1)
    return '.'.join(_v)


def read_file(f):
    try:
        with open(f, 'r') as _file:
            return _file.read()
    except Exception:
        return ''

setup(
    name='nimoy',
    version=__version__,
    url='https://github.com/laco/python-nimoy/',
    download_url='https://github.com/laco/python-nimoy/tarball/' + __version__,
    license='Apache2.0',
    author='László Andrási',
    author_email='mail@laszloandrasi.com',
    description='Python3 Database Access Layer with Simple Functions and Dictionaries',
    long_description=read_file('README.md') + '\n\n',
    packages=['nimoy', 'nimoy.backends', 'nimoy.fields'],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=['mongomock', 'pymongo', 'python-datetime2', 'PyYAML'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



from nimoy.utils import import_class
from nimoy.schemas import DBSchema
import uuid as _uuid


class DatabaseConnection(object):

    def __init__(self, **kwargs):
        backend_cls = kwargs.pop('backend', None)
        if backend_cls is None:
            raise AttributeError("Missing 'backend' parameter.")
        self.backend = import_class(backend_cls)(self)

        schema = kwargs.pop('schema', None)
        if schema is None:
            raise AttributeError("Missing 'schema' parameter.")
        elif isinstance(schema, dict):
            self.schema = DBSchema.from_dict(schema)
        elif isinstance(schema, str):
            self.schema = DBSchema.from_json(schema)
        self.options = kwargs

    def _add_version_info(self, schema_name, _data):
        _data['__sv'] = self.schema.get_fingerprint(schema_name)
        return _data

    def put_item(self, schema_name, _data, overwrite=False):
        return self.backend.put_item(
            schema_name,
            self._add_version_info(
                schema_name,
                self.backend.prepare_data(
                    schema_name,
                    self.schema.validate_data(
                        schema_name, _data))),
            overwrite=overwrite)

    def get_item(self, schema_name, _id):
        return self.backend.get_item(schema_name, _id)

    def delete_item(self, schema_name, _id):
        pass

    def query(self, schema_name, _w, limit=10):
        return self.backend.query(schema_name, _w, limit)

    def scan(self, schema_name, _w, limit=10):
        return self.backend.scan(schema_name, _w, limit)

    def query_count(self, schema_name, _w):
        return self.backend.query_count(schema_name, _w)

    def uuid(self):
        return _uuid.uuid4().hex



from nimoy.connection import DatabaseConnection


def init_adapter(cn):
    nimoy_config = cn.g_('app_config').get('nimoy_config', {})
    nimoy_schemas = cn.g_('app_config').get('nimoy_schemas', {})
    db = DatabaseConnection(schema=nimoy_schemas, **nimoy_config)
    cn.s_('nimoy_db', db)


def get_item(cn, schema_name, _id, **kw):
    return cn.g_('nimoy_db').get_item(schema_name, _id, **kw)


def put_item(cn, schema_name, _data, **kw):
    return cn.g_('nimoy_db').put_item(schema_name, _data, **kw)


def delete_item(cn, schema_name, _id, **kw):
    return cn.g_('nimoy_db').delete_item(schema_name, _id, **kw)


def query(cn, schema_name, _w, limit=10, **kw):
    return cn.g_('nimoy_db').query(schema_name, _w, limit, **kw)


def query_count(cn, schema_name, _w, **kw):
    return cn.g_('nimoy_db').query_count(schema_name, _w, **kw)


def scan(cn, schema_name, _w, limit=10, **kw):
    return cn.g_('nimoy_db').scan(schema_name, _w, limit, **kw)


def uuid(cn):
    return cn.g_('nimoy_db').uuid()
